/*
	fonctions dfinies dans Proba:
       pPoisson, fPoisson, sPoisson, &
       fnormal, pnormal, &
       fgamma, pgamma, &
       hpsort, ihpsort, iMerge1, rMerge1, locate,
	   gammp, gammq
*/

/*
	compilation: javac -target 1.1 -encoding ISO-8859-1 Proba.java
*/
public class Proba {

  /**
    * mu: Moyenne de la loi de Poisson
    * k: valeur de la v.a.
    * throws Exception if mu < 0
    */
  public static double pPoisson(double mu, int k) throws Exception {
    if (mu < 0)
		throw new Exception("pPoisson: mu < 0");
    else if (k < 0 || mu == 0 )
       return(0);
    else
       return(Math.exp((double)k * Math.log(mu) - gamln((double)(k+1)) - mu));
  }


  /**
	* Tres imprecis car on n'utilise pas la fonction Log de IGamma !!!
	*/
  public static double[] fPoisson(double mu, double[] x) {
    double xx, f2;
	double f[] = new double[x.length];

	for (int i=0; i<x.length; i++) {
       xx = Math.floor(x[i]) + 1;
       if (xx > 0)
          f[i] = gammq(xx, mu);
       else
          f[i] = 0;
    }
	return(f);
  }


  /**
   * Generates a IID Poisson sample of size n with mean mu.
   */
  public static int[] sPoisson(double mu, int n) {
  	double[] t = new double[n];
  	double[] dt = new double[n];
	int[] v = new int[n];

	for (int i=0; i<n; i++) {
		v[i] = 0;
		t[i] = 0;
	}
	boolean test = true;
	for (int i=0; i<n; i++)
		if (t[i] < mu)
			test = false;
    while (!test) {
	   for (int i=0; i<n; i++)
          dt[i] = Math.random();
	   for (int i=0; i<n; i++)
          t[i] = t[i] - Math.log(dt[i]);
	   test = true;
		for (int i=0; i<n; i++)
			if (t[i] < mu) {
				test = false;
				v[i] = v[i] + 1;
			}
    }
	return(v);
  }

  /**
   * Generates a IID Poisson sample of size 1 with mean mu.
   */
  public static int sPoisson1(double mu) {
  	double t = 0;
	int v = 0;
  	double dt;

	boolean test = (t < mu);
    while (test) {
       dt = Math.random();
       t -= Math.log(dt);
	   test = (t < mu);
		if (test)
		    v++;
    }
	return(v);
  }

    /**
      * Distribution function of a normal r.v.
      */
  public static double[] fnormal(double mu, double sigma, double[] x) {
    int n = x.length;
  	double[] y = new double[n];
  	double[] f = new double[n];

	for (int i=0; i<n; i++)
    	y[i] = (x[i]-mu)/sigma;
	for (int i=0; i<n; i++)
       f[i] = ndf(y[i]);
	 return(f);
  }

    /**
      * p.d.f. of a normal r.v.
	  */
  public static double[] pnormal(double mu, double sigma, double[] x) {
    int n = x.length;
  	double[] y = new double[n];
  	double[] p = new double[n];
    double h;

	for (int i=0; i<n; i++)
    	y[i] = (x[i] - mu) / sigma;
    h = Math.sqrt(-2.0*Math.log(Double.MIN_VALUE)); // h = Double.POSITIVE_INFINITY ?
	for (int i=0; i<n; i++)
		if (Math.abs(y[i]) <= h)
			p[i] = Math.exp(-0.5 * y[i] * y[i]);
		else
			p[i] = 0;
	for (int i=0; i<n; i++)
    	p[i] = p[i]*(1.0/Math.sqrt(2*Math.PI)/sigma);
	
	 return(p);
  }


    /**
      * Distribution function of a Gamma (Erlang) r.v.
	  */
  public static double[] fgamma(double nu, double lambda, double[] x) {
    double xx;
    int n = x.length;
  	double[] f = new double[n];

	for (int i=0; i<n; i++) {
       xx = x[i]*lambda;
       if (xx > 0)
          f[i] = gammp(nu, xx);
       else
          f[i] = 0;
    }
	/*
	for (int i=0; i<n; i++)
		if (f[i] > 0)
			f[i] = Math.exp(Math.log(f[i])-gamln(nu));
	*/
	 return(f);
  }

  /**
    * entre:
	*    nu: ordre de la loi ("Nombre de photons observes")
	*    lambda: taux d'arrive des photons.
	*    x
	* sortie: p
	*/
  public static double[] pgamma(double nu, double lambda, double[] x) throws Exception {

    int n = x.length;
  	double[] p = new double[n];
	  
    if (nu <= 0 || lambda <= 0)
		throw new Exception("pgamma: nu <= 0 || lambda <= 0");
    else {
       if (nu == 1) {
          for (int i=0; i<n; i++)
		  	if (x[i] < 0)
			  	p[i] = 0;
			else
			  	p[i] = lambda*Math.exp(-lambda*x[i]);
       } else {
          for (int i=0; i<n; i++)
		  	if (x[i] < 0)
			  	p[i] = 0;
			else
			  	p[i] = Math.exp(nu*Math.log(lambda) - gamln(nu) - lambda*x[i]) * Math.pow(x[i], nu - 1);
       }
	   return(p);
    }
  }


  /*
  	Service routines
  */

    /**
	  * Natural logarithm of the (complete) Gamma function.
	  */
  public static double gammln(double xx) {
  	final double stp=2.50662827465, fpf=5.5;
	double[] cof = {76.18009173, -86.50532033, 24.01409822, -1.231739516, .120858003E-2, -.536382E-5};
    double x, tmp, ser;

    x = xx - 1;
    tmp = x + fpf;
    tmp = (x + 0.5)*Math.log(tmp) - tmp;
    ser = 1;
    for (int i=0; i<6; i++) {
       x = x + 1;
       ser = ser + cof[i]/x;
    }
    return(tmp + Math.log(stp * ser));
  }


    /**
      *     evaluation of LN(GAMMA(A)) for positive a
      *
      *     D = 0.5*(LN(2*PI) - 1)
	  *
	  * entre: a
	  * sortie: double
	  *
	  */

  public static double gamln(double a) {
    double t, w;
    int n;

    final double d = .418938533204673;
    final double
         c0=.833333333333333E-01, c1=-.277777777760991E-02,
         c2=.793650666825390E-03, c3=-.595202931351870E-03,
         c4=.837308034031215E-0,  c5=-.165322962780713E-02;

    if (a > 0.8) {
       if (a > 2.25) { 
          if (a >= 10.0) {
             t = Math.pow(1/a, 2);
             w = (((((c5*t + c4)*t + c3)*t + c2)*t + c1)*t + c0)/a;
             return((d+w) + (a-0.5)* (Math.log(a)-1));
          } else {
             n = (int)(a - 1.25);
             t = a;
             w = 1;
             for (int i=0; i<n; i++) {
                t = t - 1;
                w = t*w;
             }
             return(gamln1(t-1) + Math.log(w));
          }
       } else {
          t = (a-0.5) - 0.5;
          return(gamln1(t));
       }
    } else
       return(gamln1(a) - Math.log(a));
  }

  /**
    * evaluation of LN(GAMMA(1 + a)) for -0.2 <= a <= 1.25
	*/
  public static double gamln1(double a) {
    double w, x;
	
    final double
         p0= .577215664901533, p1= .844203922187225,
         p2=-.168860593646662, p3=-.780427615533591,
         p4=-.402055799310489, p5=-.673562214325671E-01,
         p6=-.271935708322958E-02;
    final double
         q1=.288743195473681E+01, q2=.312755088914843E+01,
         q3=.156875193295039E+01, q4=.361951990101499E+00,
         q5=.325038868253937E-01, q6=.667465618796164E-03;
    final double
         r0=.422784335098467E+00, r1=.848044614534529E+00,
         r2=.565221050691933E+00, r3=.156513060486551E+00,
         r4=.170502484022650E-01, r5=.497958207639485E-03;
    final double
         s1=.124313399877507E+01, s2=.548042109832463E+00,
         s3=.101552187439830E+00, s4=.713309612391000E-02,
         s5=.116165475989616E-03;

    if (a >= 0.6) {
       x = (a-0.5) - 0.5;
       w = (((((r5*x + r4)*x + r3)*x + r2)*x + r1)*x + r0)/
            (((((s5*x + s4)*x + s3)*x + s2)*x + s1)*x + 1);
       return(x*w);
    } else {
       w = ((((((p6*a + p5)*a + p4)*a + p3)*a + p2)*a + p1)*a + p0)/
            ((((((q6*a + q5)*a + q4)*a + q3)*a + q2)*a + q1)*a + 1);
       return(-a*w);
    }
  }

  /*
        Sort/Merge Programs
  */

    /**
	  *	Indexes an array arr[0:n-1] i.e., outputs the array indx[0:n-1] such that 
      * arr[indx[j]] is in ascending order for j=0,1,...,n-1.
      * 
	  * entre: n, arr
	  * sortie: indx
	  */

  public static void indexx(int n, double[] arr, int[] indx) {

    final int M=7, NSTACK=50;

    int i, indxt, ir=n, itemp, k, l=1, j;
	int jstack=0;
    int[] istack = new int[NSTACK+1];
    double a;

	for (j=1;j<=n;j++) indx[j]=j;
	for (;;) {
		if (ir-l < M) {
			for (j=l+1;j<=ir;j++) {
				indxt=indx[j];
				a=arr[indxt];
				for (i=j-1;i>=1;i--) {
					if (arr[indx[i]] <= a) break;
					indx[i+1]=indx[i];
				}
				indx[i+1]=indxt;
			}
			if (jstack == 0) break;
			ir=istack[jstack--];
			l=istack[jstack--];
		} else {
			k=(l+ir) >> 1;
			itemp = indx[k];
			indx[k] = indx[l+1];
			indx[l+1] = itemp;
			if (arr[indx[l+1]] > arr[indx[ir]]) {
				itemp = indx[l+1];
				indx[l+1] = indx[ir];
				indx[ir] = itemp;
			}
			if (arr[indx[l]] > arr[indx[ir]]) {
				itemp = indx[l];
				indx[l] = indx[ir];
				indx[ir] = itemp;
			}
			if (arr[indx[l+1]] > arr[indx[l]]) {
				itemp = indx[l+1];
				indx[l+1] = indx[l];
				indx[l] = itemp;
			}
			i=l+1;
			j=ir;
			indxt=indx[l];
			a=arr[indxt];
			for (;;) {
				do i++; while (arr[indx[i]] < a);
				do j--; while (arr[indx[j]] > a);
				if (j < i) break;
				itemp = indx[i];
				indx[i] = indx[j];
				indx[j] = itemp;
			}
			indx[l]=indx[j];
			indx[j]=indxt;
			jstack += 2;
			if (jstack > NSTACK)
				System.err.println("NSTACK too small in indexx.");
			if (ir-i+1 >= j-l) {
				istack[jstack]=ir;
				istack[jstack-1]=i;
				ir=j-1;
			} else {
				istack[jstack]=j-1;
				istack[jstack-1]=l;
				l=i;
			}
		}
	}
}

    /**
      *  Sort *double* array ra in ascending order
	  * entre-sortie: ra
      */
  public static void hpsort(double[] ra) {
	int i,ir,j,l,n;
	double rra;

	n = ra.length;
	if (n < 2) return;
	l=(n >> 1);
	ir=n-1;
	for (;;) {
		if (l > 0) {
			rra=ra[--l];
		} else {
			rra=ra[ir];
			ra[ir]=ra[0];
			if (--ir == 0) {
				ra[0]=rra;
				break;
			}
		}
		i=l;
		j=l+l+1;
		while (j <= ir) {
			if (j < ir && ra[j] < ra[j+1]) j++;
			if (rra < ra[j]) {
				ra[i]=ra[j];
				i=j;
				j = (j << 1)+1;
			} else j=ir+1;
		}
		ra[i]=rra;
	}
  }

    /**
      *  Sort *int* array ra in ascending order
	  * entre-sortie: ra
      */
  public static void hpsort(int[] ra) {
	int i,ir,j,l,n;
	int rra;

	n = ra.length;
	if (n < 2) return;
	l=(n >> 1);
	ir=n-1;
	for (;;) {
		if (l > 0) {
			rra=ra[--l];
		} else {
			rra=ra[ir];
			ra[ir]=ra[0];
			if (--ir == 0) {
				ra[0]=rra;
				break;
			}
		}
		i=l;
		j=l+l+1;
		while (j <= ir) {
			if (j < ir && ra[j] < ra[j+1]) j++;
			if (rra < ra[j]) {
				ra[i]=ra[j];
				i=j;
				j = (j << 1)+1;
			} else j=ir+1;
		}
		ra[i]=rra;
	}
  }


  /**
    * Merge last element of otherwise sorted integer array a
	* entre-sortie: a
	*/
  public static void iMerge1(int[] a) {
  	iMerge1(a, a.length);
  }
  
  /**
    * Merge last element of otherwise sorted integer array a[0:na]
	* entre-sortie: a
	*/
  public static void iMerge1(int[] a, int na) {
    int inotsup, isave;

    if (na<=1)
		return;

    inotsup = iLocate(a, na);

    // Do the right shift to make space for a[na-1]
    isave = a[na-1];
	for (int i=na-1; i>=inotsup+2; i--)
		a[i] = a[i-1];
    a[inotsup+1] = isave;
  }

  /**
    * Merge last element of otherwise sorted double array a.
	* entre-sortie: a
	*/
  public static void rMerge1(double[] a) {
  	rMerge1(a, a.length);
  }
  
  /**
    * Merge last element of otherwise sorted double array a[0:na-1]
	* entre-sortie: a
	*/
  public static void rMerge1(double[] a, int na) {
    int inotsup;
	double isave;

    if (na<=1)
		return;

    inotsup = rLocate(a[na-1], a, na-1);

    // Do the right shift to make space for a[na-1]
    isave = a[na-1];
	for (int i=na-1; i>=inotsup+2; i--)
		a[i] = a[i-1];
    a[inotsup+1] = isave;
  }


    /**
	  * Locate returns rank of greatest a[0:n-1] not greater than x.
      * a[0:n-1] _must_ be sorted in ascending order.
	  */
  public static int rLocate(double x, double[] a, int na) {
    int imin, imax, itest;

    imin = -1;
    imax = na;
    for (;;) {
       itest = (int)Math.round(0.5*(imin+imax));
       if (x > a[itest])
          imin = itest;
       else
          imax = itest;
       if (imax-imin <= 1)
	   		break;
    }
    return(Math.min(imin, imax));
  }


    /**
	  * Locate returns rank of greatest a[0:n-2] not greater than a[n-1].
      * a[0:n-2] _must_ be sorted in ascending order.
	  */
  public static int iLocate(int[] a) {
  	return(iLocate(a, a.length));
  }
  
    /**
	  * Locate returns rank of greatest a[0:n-2] not greater than a[n-1].
      * a[0:n-2] _must_ be sorted in ascending order.
	  */
  public static int iLocate(int[] a, int n) {
  	int x = a[n-1];
    int na, imin, imax, itest;

    na = n-1;
    imin = -1;
    imax = na;
    for (;;) {
       itest = (int)Math.round(0.5*(imin+imax));
       if (x > a[itest])
          imin = itest;
       else
          imax = itest;
       if (imax-imin <= 1)
	   		break;
    }
    return(Math.min(imin, imax));
  }

/**
  * ndf returns the normal distribution function
  */
  public static double ndf(double x) {
    return(erf(x/Math.sqrt(2))/2 + 0.5);
  }

/**
  * error function erf(x)
  */
  public static double erf(double x) {
    if (x < 0)
      return(-gammp(0.5, Math.pow(x,2)));
	else
	  return(gammp(0.5, Math.pow(x,2)));
  }

	/**
	  * gammln pour gammp... semble diffrente de l'autre gammln... normalise ?
	  *
	  * Returns the value ln[G(x)] for x > 0.
	  * Internal arithmetic will be done in double precision, a nicety that you can omit if five-figure
	  * accuracy is good enough.
	  */
	public static double gammln2(double x) {
		final double[] cof = {76.18009172947146e0, -86.50532032941677e0,
			24.01409824083091e0, -1.231739572450155e0,  .1208650973866179e-2,
			-.5395239384953e-5};
		final double stp = 2.5066282746310005e0;
		final double fpf = 5.5;
		double ser, tmp, y;
		
		y = x;
		tmp = x + fpf;
		tmp = (x + 0.5)*Math.log(tmp) - tmp;
		ser = 1.000000000190015;
		for (int i=0; i<6; i++) {
			y = y + 1.;
			ser = ser + cof[i]/y;
		}
		return(tmp + Math.log(stp*ser/x));
	}
	
    /**
	  * Returns the incomplete gamma function P(a; x).
	  */
  public static double gammp(double a, double x) {
// USES gcf,gser
	if (x < 0 || a <= 0)
		System.err.println("bad arguments in gammp");
	if (x < a + 1)
		return(gser(a, x));
	else
		return(1 - gcf(a, x));
  }

  /**
    * Returns the incomplete gamma function Q(a; x) = 1 - P(a; x).
	*/
  public static double gammq(double a, double x) {
// USES gcf,gser
	if (x < 0 || a <= 0)
		System.err.println("bad arguments in gammq");
	if (x < a + 1)
		return(1 - gser(a, x));
	else
		return(gcf(a, x));
  }

  /**
	* Returns the incomplete gamma function P(a; x) evaluated by its series representation
	*/
  public static double gser(double a, double x) {
	final int ITMAX = 100;
	final double EPS = 3.e-7;
	double gln;
// USES gammln
	double ap, del, sum;
	gln = gammln2(a);
	if (x <= 0) {
		if (x < 0)
			System.err.println("x < 0 in gser");
		return(0);
	}
	ap = a;
	sum = 1./a;
	del = sum;
	boolean problem = true;
	for (int i=0; i<ITMAX; i++) {
		ap = ap + 1.;
		del = del * x / ap;
		sum = sum + del;
		if (Math.abs(del) < Math.abs(sum)*EPS) {
			problem = false;
			break;
		}
	}
	if (problem)
		System.err.println("a too large, ITMAX too small in gser");
	return(sum * Math.exp(-x + a * Math.log(x) - gln));
  }

  /**
	 * Returns the incomplete gamma function Q(a; x) evaluated by its continued fraction representation
	 * as gammcf.
	 * Parameters: ITMAX is the maximum allowed number of iterations; EPS is the relative accuracy;
	 * FPMIN is a number near the smallest representable floating-point number.
	 */
  public static double gcf(double a, double x) {
	final int ITMAX = 100;
	final double EPS=3.e-7;
	final double FPMIN=1.e-30;
	double gln;
// USES gammln
	double an,b,c,d,del,h;
	gln = gammln2(a);
	b = x + 1. - a;
	c = 1./FPMIN;
	d = 1./b;
	h = d;
	boolean problem = true;
	for (int i=1; i<ITMAX; i++) {
		an = -i *(i - a);
		b = b + 2.;
		d = an*d + b;
		if (Math.abs(d) < FPMIN)
			d = FPMIN;
		c = b + an/c;
		if (Math.abs(c) < FPMIN)
			c = FPMIN;
		d = 1./d;
		del = d*c;
		h = h*del;
		if (Math.abs(del-1.) < EPS) {
			problem = false;
			break;
		}
	}
	if (problem)
		System.err.println("a too large, ITMAX too small in gcf");
	return(Math.exp(-x + a*Math.log(x) - gln) * h);
  }
  
}
